#include <media/AudioTrack.h>
#include <utils/Log.h>
#include <AudioFlinger.h>
#include <MediaPlayerInterface.h>
#include <MediaPlayerService.h>
#include "audioflinger_wrapper.h"
#include <AudioSystem.h>

//----------------------------------------------------------------------------------------------------
#include <utils/Log.h>
#undef  LOG_TAG
#define LOG_TAG "HoFFPLAYAudio"
#if 1
#define gprintf(fmt, args...) LOGE("%s(%d): " fmt, __FUNCTION__, __LINE__, ##args)
#else
#define gprintf(fmt, args...)
#endif
//----------------------------------------------------------------------------------------------------

using namespace android;

typedef struct _AudioFlingerDevice {
    AudioTrack     *audio_track;
    bool            init;
                    sp < MediaPlayerBase::AudioSink > audio_sink;
} AudioFlingerDevice;

/* commonly used macro */
#define AUDIO_FLINGER_DEVICE(handle) ((AudioFlingerDevice*)handle)
#define AUDIO_FLINGER_DEVICE_TRACK(handle) \
    (AUDIO_FLINGER_DEVICE(handle)->audio_track)
#define AUDIO_FLINGER_DEVICE_SINK(handle) \
    (AUDIO_FLINGER_DEVICE(handle)->audio_sink)

AudioFlingerDeviceHandle audioflinger_device_create() {
    AudioFlingerDevice *audiodev = NULL;
    AudioTrack     *audiotr = NULL;

    // create a new instance of AudioFlinger 
    audiodev = new AudioFlingerDevice;
    if (audiodev == NULL) {
        gprintf("Error to create AudioFlingerDevice\n");
        return NULL;
    }

    // create AudioTrack
    audiotr = new AudioTrack();
    if (audiotr == NULL) {
        gprintf("Error to create AudioTrack\n");
        return NULL;
    }

    audiodev->init = false;
    audiodev->audio_track = (AudioTrack *) audiotr;
    audiodev->audio_sink = 0;
    gprintf("Create AudioTrack successfully\n");

    return (AudioFlingerDeviceHandle) audiodev;
}

AudioFlingerDeviceHandle audioflinger_device_open(void *audio_sink) {
    AudioFlingerDevice *audiodev = NULL;

    // audio_sink shall be an MediaPlayerBase::AudioSink instance
    if (audio_sink == NULL)
        return NULL;

    // create a new instance of AudioFlinger 
    audiodev = new AudioFlingerDevice;
    if (audiodev == NULL) {
        gprintf("Error to create AudioFlingerDevice\n");
        return NULL;
    }

    // set AudioSink
    audiodev->audio_track = NULL;
    audiodev->audio_sink = (MediaPlayerBase::AudioSink *) audio_sink;
    audiodev->init = false;
    gprintf("Open AudioSink successfully\n");
    return (AudioFlingerDeviceHandle) audiodev;
}

int
audioflinger_device_set(AudioFlingerDeviceHandle handle, int streamType, int channelCount, uint32_t sampleRate, int bufferCount) {
    status_t        status = NO_ERROR;

    int             format = AUDIO_FORMAT_PCM_16_BIT;

    if (handle == NULL)
        return -1;

    if (AUDIO_FLINGER_DEVICE_TRACK(handle)) {
        // bufferCount is not the number of internal buffer, but the internal
        // buffer size 
        //status = AUDIO_FLINGER_DEVICE_TRACK(handle)->set(streamType, sampleRate, format, channelCount, bufferCount);
        //status = AUDIO_FLINGER_DEVICE_TRACK(handle)->set(streamType, sampleRate, format, CHANNEL_OUT_FRONT_LEFT|CHANNEL_OUT_FRONT_RIGHT, bufferCount);
        status = AUDIO_FLINGER_DEVICE_TRACK(handle)->set(streamType, sampleRate, format, AUDIO_CHANNEL_OUT_STEREO, bufferCount);
        gprintf("Set AudioTrack, status: %d, streamType: %d, sampleRate: %d, " "channelCount: %d, bufferCount: %d\n", status, streamType, sampleRate, channelCount, bufferCount);
    }
    else if (AUDIO_FLINGER_DEVICE_SINK(handle).get()) {
        status = AUDIO_FLINGER_DEVICE_SINK(handle)->open(sampleRate, channelCount, format, bufferCount);
        gprintf("Set AudioSink, status: %d, streamType: %d, sampleRate: %d," "channelCount: %d, bufferCount: %d\n", status, streamType, sampleRate, channelCount, bufferCount);
    }

    if (status != NO_ERROR)
        return -1;

    AUDIO_FLINGER_DEVICE(handle)->init = true;

    return 0;
}

void audioflinger_device_release(AudioFlingerDeviceHandle handle) {
    if (handle == NULL)
        return;

    gprintf("Enter\n");
    if (AUDIO_FLINGER_DEVICE_TRACK(handle)) {
        gprintf("Release AudioTrack\n");
        delete          AUDIO_FLINGER_DEVICE_TRACK(handle);
    }


    if (AUDIO_FLINGER_DEVICE_SINK(handle).get()) {
        gprintf("Release AudioSink\n");
        AUDIO_FLINGER_DEVICE_SINK(handle).clear();
    }

    delete          AUDIO_FLINGER_DEVICE(handle);
}


void audioflinger_device_start(AudioFlingerDeviceHandle handle) {
    if (handle == NULL || AUDIO_FLINGER_DEVICE(handle)->init == false)
        return;

    if (AUDIO_FLINGER_DEVICE_TRACK(handle)) {
        AUDIO_FLINGER_DEVICE_TRACK(handle)->start();
    }
    else {
        AUDIO_FLINGER_DEVICE_SINK(handle)->start();
    }
}

void audioflinger_device_stop(AudioFlingerDeviceHandle handle) {
    if (handle == NULL || AUDIO_FLINGER_DEVICE(handle)->init == false)
        return;

    if (AUDIO_FLINGER_DEVICE_TRACK(handle)) {
        AUDIO_FLINGER_DEVICE_TRACK(handle)->stop();
    }
    else {
        AUDIO_FLINGER_DEVICE_SINK(handle)->stop();
    }
}

void audioflinger_device_flush(AudioFlingerDeviceHandle handle) {
    if (handle == NULL || AUDIO_FLINGER_DEVICE(handle)->init == false)
        return;

    if (AUDIO_FLINGER_DEVICE_TRACK(handle)) {
        AUDIO_FLINGER_DEVICE_TRACK(handle)->flush();
    }
    else {
        AUDIO_FLINGER_DEVICE_SINK(handle)->flush();
    }
}

void audioflinger_device_pause(AudioFlingerDeviceHandle handle) {
    if (handle == NULL || AUDIO_FLINGER_DEVICE(handle)->init == false)
        return;

    if (AUDIO_FLINGER_DEVICE_TRACK(handle)) {
        AUDIO_FLINGER_DEVICE_TRACK(handle)->pause();
    }
    else {
        AUDIO_FLINGER_DEVICE_SINK(handle)->pause();
    }
}

void audioflinger_device_mute(AudioFlingerDeviceHandle handle, int mute) {
    if (handle == NULL || AUDIO_FLINGER_DEVICE(handle)->init == false)
        return;

    if (AUDIO_FLINGER_DEVICE_TRACK(handle)) {
        AUDIO_FLINGER_DEVICE_TRACK(handle)->mute((bool) mute);
    }
    else {
        // do nothing here, because the volume/mute is set in media service layer
    }
}

int audioflinger_device_muted(AudioFlingerDeviceHandle handle) {
    if (handle == NULL || AUDIO_FLINGER_DEVICE(handle)->init == false)
        return -1;

    if (AUDIO_FLINGER_DEVICE_TRACK(handle)) {
        return (int) AUDIO_FLINGER_DEVICE_TRACK(handle)->muted();
    }
    else {
        // do nothing here, because the volume/mute is set in media service layer
        return -1;
    }
}


void audioflinger_device_set_volume(AudioFlingerDeviceHandle handle, float left, float right) {
    if (handle == NULL || AUDIO_FLINGER_DEVICE(handle)->init == false)
        return;

    if (AUDIO_FLINGER_DEVICE_TRACK(handle)) {
        AUDIO_FLINGER_DEVICE_TRACK(handle)->setVolume(left, right);
    }
    else {
        // do nothing here, because the volume/mute is set in media service layer
        return;
    }
}

ssize_t audioflinger_device_write(AudioFlingerDeviceHandle handle, const void *buffer, size_t size) {
    if (handle == NULL || AUDIO_FLINGER_DEVICE(handle)->init == false)
        return -1;
    if (AUDIO_FLINGER_DEVICE_TRACK(handle)) {
        return AUDIO_FLINGER_DEVICE_TRACK(handle)->write(buffer, size);
    }
    else {
        return AUDIO_FLINGER_DEVICE_SINK(handle)->write(buffer, size);
    }
}

int audioflinger_device_frameCount(AudioFlingerDeviceHandle handle) {
    if (handle == NULL || AUDIO_FLINGER_DEVICE(handle)->init == false)
        return -1;
    if (AUDIO_FLINGER_DEVICE_TRACK(handle)) {
        return (int) AUDIO_FLINGER_DEVICE_TRACK(handle)->frameCount();
    }
    else {
        return (int) AUDIO_FLINGER_DEVICE_SINK(handle)->frameCount();
    }
}

int audioflinger_device_frameSize(AudioFlingerDeviceHandle handle) {
    if (handle == NULL || AUDIO_FLINGER_DEVICE(handle)->init == false)
        return -1;
    if (AUDIO_FLINGER_DEVICE_TRACK(handle)) {
        return (int) AUDIO_FLINGER_DEVICE_TRACK(handle)->frameSize();
    }
    else {
        return (int) AUDIO_FLINGER_DEVICE_SINK(handle)->frameSize();
    }
}

int64_t audioflinger_device_latency(AudioFlingerDeviceHandle handle) {
    if (handle == NULL || AUDIO_FLINGER_DEVICE(handle)->init == false)
        return -1;
    if (AUDIO_FLINGER_DEVICE_TRACK(handle)) {
        return (int64_t) AUDIO_FLINGER_DEVICE_TRACK(handle)->latency();
    }
    else {
        return (int64_t) AUDIO_FLINGER_DEVICE_SINK(handle)->latency();
    }
}

int audioflinger_device_format(AudioFlingerDeviceHandle handle) {
    if (handle == NULL || AUDIO_FLINGER_DEVICE(handle)->init == false)
        return -1;
    if (AUDIO_FLINGER_DEVICE_TRACK(handle)) {
        return (int) AUDIO_FLINGER_DEVICE_TRACK(handle)->format();
    }
    else {
        // do nothing here, MediaPlayerBase::AudioSink doesn't provide format()
        // interface
        return -1;
    }
}

int audioflinger_device_channelCount(AudioFlingerDeviceHandle handle) {
    if (handle == NULL || AUDIO_FLINGER_DEVICE(handle)->init == false)
        return -1;
    if (AUDIO_FLINGER_DEVICE_TRACK(handle)) {
        return (int) AUDIO_FLINGER_DEVICE_TRACK(handle)->channelCount();
    }
    else {
        return (int) AUDIO_FLINGER_DEVICE_SINK(handle)->channelCount();
    }
}

uint32_t audioflinger_device_sampleRate(AudioFlingerDeviceHandle handle) {
    if (handle == NULL || AUDIO_FLINGER_DEVICE(handle)->init == false)
        return 0;
    if (AUDIO_FLINGER_DEVICE_TRACK(handle)) {
        return (int) AUDIO_FLINGER_DEVICE_TRACK(handle)->getSampleRate(); // ghcstop fix for donut
    }
    else {
        // do nothing here, MediaPlayerBase::AudioSink doesn't provide sampleRate()
        // interface
        return -1;
    }
}
