#include "surfaceflinger_wrapper.h"

#include <surfaceflinger/Surface.h>
#include <surfaceflinger/ISurfaceComposer.h>
#include <ui/GraphicBufferMapper.h>
#include <gui/ISurfaceTexture.h>
#include <media/stagefright/ColorConverter.h>
#include <gui/ISurfaceTexture.h>
#include <gui/SurfaceTexture.h>
#include <gui/SurfaceTextureClient.h>

#include <system/window.h>
#include <android/native_window.h>
#include <sec_format.h>
#include <SEC_OMX_Def.h>

using namespace android;

//-----------------------------------------------------------------------------------
#include <utils/Log.h>
#undef  LOG_TAG
#define LOG_TAG "HoFFPLAYRenderer"
#if 1
#define gprintf(fmt, args...) LOGE("%s(%d): " fmt, __FUNCTION__, __LINE__, ##args)
#else
#define gprintf(fmt, args...)
#endif
//-----------------------------------------------------------------------------------

OMX_COLOR_FORMATTYPE eColorFormat;
ColorConverter* mConverter;

sp<ANativeWindow> mNativeWindow;
int32_t mWidth, mHeight;
int32_t mCropLeft, mCropTop, mCropRight, mCropBottom;
int32_t mCropWidth, mCropHeight;

// ȭ ũ   ϴ κ
int hovideo_prepare(int left, int top, int right, int bottom, int width, int height, int ColorFormat) {
    status_t err;
    uint32_t usage = 0;
    uint32_t nBufferCountActual = 4;

    mWidth = width;
    mHeight = height;

    android_native_rect_t crop;
    crop.left = 0;
    crop.top = 0;
    crop.right = mWidth;
    crop.bottom = mHeight;

    native_window_set_crop(mNativeWindow.get(), &crop);
    err = native_window_set_scaling_mode(mNativeWindow.get(), NATIVE_WINDOW_SCALING_MODE_SCALE_TO_WINDOW);

    if (err != OK) {
        return -1;
    }

    //⿡ YUV420P ұ  SW Converter YUV420P ȯ ̴.
    eColorFormat = (OMX_COLOR_FORMATTYPE)HAL_PIXEL_FORMAT_YCbCr_420_P;
    
    err = native_window_set_buffers_geometry(mNativeWindow.get(), width, height, eColorFormat);
    if (err != 0) {
        gprintf("native_window_set_buffers_geometry failed: %s (%d)", strerror(-err), -err);
        return err;
    }
    if (usage & GRALLOC_USAGE_PROTECTED) {
        int queuesToNativeWindow = 0;
        err = mNativeWindow->query(
                mNativeWindow.get(), NATIVE_WINDOW_QUEUES_TO_WINDOW_COMPOSER,
                &queuesToNativeWindow);
        if (err != 0) {
            gprintf("error authenticating native window: %d", err);
            return err;
        }
        if (queuesToNativeWindow != 1) {
            gprintf("native window could not be authenticated");
            return PERMISSION_DENIED;
        }
    }

    err = native_window_set_usage(
            mNativeWindow.get(), usage | GRALLOC_USAGE_SW_WRITE_OFTEN);

    if (err != 0) {
        gprintf("native_window_set_usage failed: %s (%d)", strerror(-err), -err);
        return -1;
    }

    int minUndequeuedBufs = 0;
    err = mNativeWindow->query(mNativeWindow.get(),
            NATIVE_WINDOW_MIN_UNDEQUEUED_BUFFERS, &minUndequeuedBufs);
    if (err != 0) {
        gprintf("NATIVE_WINDOW_MIN_UNDEQUEUED_BUFFERS query failed: %s (%d)",
                strerror(-err), -err);
        return err;
    }

    err = native_window_set_buffer_count(mNativeWindow.get(), nBufferCountActual);
    if (err != 0) {
        gprintf("native_window_set_buffer_count failed: %s (%d)", strerror(-err), -err);
        return err;
    }
    return 0;
}

//JAVA Application Ȯ Surface ޹޾ Native ÷̾  ϴ κ̴.
int hovideo_setVideoSurface(void* surfaceTexture) {
    int ret = 0;
    mNativeWindow = static_cast<ANativeWindow *>(surfaceTexture);
    return 0;
}

//ÿ  ش.
int hovideo_release(void) {
    status_t err = native_window_api_disconnect(mNativeWindow.get(), NATIVE_WINDOW_API_MEDIA);

    if (err != OK) {
        gprintf("native_window_api_disconnect returned an error: %s (%d)", strerror(-err), err);
    }
    return 0;
}

//ڵ Ϸ ͸ Surface ޸𸮿 Ͽ ȭ Ѵ.
int hovideo_post(char* yData, char* uData, char* vData) {
    status_t error;
    status_t err;
    uint32_t usage = 0;
    uint32_t format = 0;
    uint32_t FrameSize = mWidth*mHeight;

    int64_t timeUs = 0;

    ANativeWindowBuffer *buf;

    if ((err = mNativeWindow->dequeueBuffer(mNativeWindow.get(), &buf)) != 0) {
        gprintf("Surface::dequeueBuffer returned error %d", err);
        return -1;
    }

    if ( mNativeWindow->lockBuffer(mNativeWindow.get(), buf) != 0 ) {
        return -1;
    }

    GraphicBufferMapper &mapper = GraphicBufferMapper::get();
    Rect bounds(mWidth, mHeight);
    void *dst;
    unsigned char *temp_ptr;
    if (mapper.lock(buf->handle, GRALLOC_USAGE_SW_WRITE_OFTEN, bounds, &dst) != 0) {
        return -1;
    }

    temp_ptr = (unsigned char *)dst;

    memcpy(temp_ptr, yData, mWidth * mHeight);
    memcpy(temp_ptr+FrameSize, uData, mWidth * mHeight / 4);
    memcpy(temp_ptr+FrameSize+(mWidth * mHeight / 4), vData, mWidth * mHeight / 4);

    if (mapper.unlock(buf->handle) != 0) {
        return -1;
    }

    if ((err = mNativeWindow->queueBuffer(mNativeWindow.get(), buf)) != 0) {
        gprintf("Surface::queueBuffer returned error %d", err);
    }

    native_window_set_buffers_timestamp(mNativeWindow.get(), timeUs * 1000);
    return 0;
}

